package controller;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import champ2010client.Action;
import champ2010client.Controller;
import champ2010client.SensorModel;

/**
 * 1.0	Kontroler vozidla zatacajuci podla najdlhsieho senzora. Funkcionalitou je
 * 		viac-menej totozny s Marianovym z OdmerajSmyk.
 * 		(Ivan)
 */
public class Prototype extends Controller
{	
	
	//instancia recovery modul
	RecoveryModule recMod;
	//radenie
	//zoradeny rad otacok za poslenych 30 tikov
	private Deque<Double> rpm30Tick = new ArrayDeque<Double>(); 
	private double clutchCur=0;
	boolean RPMGrows=false;
	boolean notOnTrack =false;
	int dontGear =0;// 7 tikov trva kym otacky zacnu po radeni stupat
	/*
	 * Koeficient frikcie medzi pneumatikou a vozovkou; tato hodnota je urcena na zaklade brzdnej drahy pri 
	 * styroch roznych rychlostiach na okruhu CG speedway, na dalsich troch "Road" tratiach odskusana -> na
	 * Road tratiach je mozne ju pouzivat zatial
	 */
	final float mi=(float)34;

	// rad sluziaci na komunikaciu s vlaknom vytvarajucim model
	private final BlockingQueue<TickInfo> tickInfoPipe;
	
	// rad sluziaci na komunikaciu s vlaknom vytvarajucim model
	private final BlockingQueue<SensorInfo> sensorInfoPipe;

	//premenna pre sirku trate
	private double trackWidth=0;
	
	public Prototype()
	{
		tickInfoPipe = new LinkedBlockingQueue<TickInfo>(1000);
		
		Thread t = new Thread(new TurnDetector(tickInfoPipe));
		t.setDaemon(true);
		t.setPriority(Thread.MIN_PRIORITY);
		t.start();
		
		sensorInfoPipe = new LinkedBlockingQueue<SensorInfo>(1000);
		
		Thread t2 = new Thread(new SensorDetector(sensorInfoPipe));
		t2.setDaemon(true);
		t2.setPriority(Thread.MIN_PRIORITY);
		t2.start();
		
		recMod=new RecoveryModule();

	}

	public Action control(SensorModel sm)
	{
		Action action = new Action ();
		
		//zistenie sirky trate
		trackWidth=sm.getTrackEdgeSensors()[0]+sm.getTrackEdgeSensors()[18]; 
		//System.out.println(trackWidth+"sirka trate");
		
		//updatne sa zoznam otacok za poslednych 30 tikov
		if(rpm30Tick.size()==15 && dontGear==0)
			rpm30Tick.pollFirst();
		if(dontGear==0)
			rpm30Tick.addLast(sm.getRPM());
		else
			dontGear--;
			
		/*if(notOnTrack){
			notOnTrack=false;
			RPMGrows=false;
		}else			
			RPMGrows= isDequeGrowing(rpm30Tick);*/
		
		
		// vyletel som na lavu stranu cesty alebo bol v recovrery state left a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if (sm.getTrackPosition()>1 || recMod.isInRecoveryStateLeft()){
			notOnTrack=true;
			return recMod.recoverLeft(sm);
		}
		
		// vyletel som na pravu stranu cesty alebo bol v recovrery state right a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if (sm.getTrackPosition()<-1 || recMod.isInRecoveryStateRight()){
			notOnTrack=true;
			return recMod.recoverRight(sm);
		}
		
		
		/*
		 * rychlost - akceleracia naplno do dosiahnutia maximalnej rychlosti, ktoru je mozne ubrzdit na vzdialenost
		 * najdlhsieho senzora,
		 * vzorec			v = odmocnina(2 * g * mi * brzdna draha)
		 */
		if (sm.getSpeed ()< Math.sqrt(2 * 9.81 * mi * sm.getTrackEdgeSensors()[18 - (int) ((computeAvgLongestSensorAngle(sm.getTrackEdgeSensors())+90)/10)]))
		{	action.accelerate = ESP.filterASR(sm, ESP.actionAccelAktual);
			ESP.actionBrakeAktual = 0;
			ESP.actionBrakeTickBack = 0;
		}
		else{
			action.brake = ESP.filterABS(sm, ESP.actionBrakeAktual);
			ESP.actionAccelAktual = 0;
			ESP.actionAccelTickBack = 0;
		}
	
		// * riadenie *
		// zmena smeru na zaklade najdlhsieho senzora
		action.steering = steeringAdjustment(computeAvgLongestSensorAngle(sm.getTrackEdgeSensors()));
		
		
		/* testovanie navigacie na konkretne miesto - nastavene na prvu zakrutu trate CG SpeedWay
		if(sm.getDistanceFromStartLine()>310) action.steering = steeringAdjustment(computeAvgLongestSensorAngle(sm.getTrackEdgeSensors()));
		if(sm.getDistanceFromStartLine()<30 && sm.getDistanceFromStartLine()>10) {
			action.steering=-0.15; //vytocenie na pravu stranu
		}
		if(sm.getDistanceFromStartLine()>30 && sm.getDistanceFromStartLine()<310) { //navigacia az po zaciatok zakruty
			System.out.println("navigujem...");
			action=navigateToPosition(action, sm, 300, 0.1); //posledny parameter urcuje vertikalnu poziciu kde mozme auto smerovat
		}*/
		
				
		// * prevodovka *
		action = getGear(sm,action,rpm30Tick);

		// posielanie informacie potrebnej na vytvaranie modelu trate
/*		try {
			tickInfoPipe.put(new TickInfo(action, sm));
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri posielani akcie do modelu trate.");
			e.printStackTrace();
		}
*/
		try
		{
			sensorInfoPipe.put(new SensorInfo((int)(computeAvgLongestSensorAngle(sm.getTrackEdgeSensors())+90)/10, 1, sm.getDistanceFromStartLine()));
		}catch (InterruptedException e) {
			System.err.println("Error: chyba pri posielani akcie do modelu trate.");
			e.printStackTrace();
		}
			
		return action;
	}
	/**
	 * metoda modulu riadenia na urcenie prevodoveho stupna s otacok motora a podla toho ci rastu/klesaju
	 * 
	 * @param sensors senzormodel
	 * @param action akcia na vykonanie
	 * @param rpmGrows boolena hovorici o tom ci otacky klesaju alebo stupaju
	 * @return akciu, ktora ma nastaveny aj prevodovy stupen 
	 */
	private Action getGear(SensorModel sensors,Action action, Deque<Double> rad){
		int gearCur;
		double rpm = sensors.getRPM();
		int[]  gearUp = {9500, 9500, 9500, 9500, 9500, 0};
		int[]  gearDown =  {0, 4300, 6200, 7000, 7300, 7700};
		gearCur=sensors.getGear();//berem akt stupen
		//som na starte a pretek este nebol odstarovany radim 1, stlacim spojku, pridam plyn 
		if(gearCur<1){
			gearCur=1;
			action.gear=gearCur;
			action.accelerate=1;
			clutchCur=1;
			action.clutch=clutchCur;
		}
		//uz je odstartovane a mam stlacenu spojku, yniyim stlacenie o 0.05, stale pridavam
		else if(clutchCur>0.65){
			action.accelerate=1;
			clutchCur-=0.05;
			action.clutch=clutchCur;
			//System.out.println("som tu");
		}
		// rychlost je mensia ako 6 sucastne su otaky vecsie/rovne otackam ked radim, a sucasne otacky v predchadzajucich 30 tikoch stupali => radim hore
		else if (gearCur < 6 && rpm >= gearUp[gearCur-1] && dontGear==0)
		{		
			if(isDequeGrowing(rad)){
				dontGear=8;
				gearCur++;
			}
		}
		// rychlost je vetsia ako 1 sucastne su otaky mensie/rovne otackam ked podradzujem a otacky za poslednych 30 tikov klesaju
		else if (gearCur > 1 && rpm <= gearDown[gearCur-1]  && dontGear==0)
		{
			if(!isDequeGrowing(rad)){
				dontGear=8;
				gearCur--;	
			}
			
		}
		if(clutchCur>0 && gearCur>1)
			clutchCur-=0.05;
		action.gear=gearCur;
		action.clutch=clutchCur;
		return action;
	}
	/**
	 * metoda vrati true ak rad<double> apon s polovice rastie a false ak nsatane opacne situacia
	 * 
	 * @param que rad cisel ktorych rast sa ma urcit
	 * @return true alebo false
	 */
	private boolean isDequeGrowing(Deque<Double> que){
		Deque<Double> pom = new ArrayDeque<Double>(que);
		boolean rastie[] = new boolean[pom.size()];
		double curValue = pom.pollFirst().doubleValue();
		//System.out.print(pom.size()+" "+curValue+ " ");
		for(int i=0;i<que.size()-1;i++){
			rastie [i]= curValue<pom.peekFirst().doubleValue();
			curValue = pom.pollFirst().doubleValue();
			//System.out.print(curValue+ " "+rastie [i]+";");
		}
		
		int h=0;
		for(int i=0;i<rastie.length;i++){
			if(rastie[i])
				h++;
		}
		//System.out.print(h+" "+pom.size());
		//System.out.println();
		if(h>=rastie.length/2)
			return true;
		else
			return false;
	}
	
	
	/**
	 * Metoda, ktora vypocita uhol medzi aktualnou poziciou auta a bodom na trati pred autom.
	 * @param sm senzory
	 * @param positionDistance vzdialenost sledovaneho bodu od startu
	 * @param position vertikalna pozicia sledovaneho bodu na trati
	 * @return uhol
	 * @author Juraj Kosmel
	 */
	private double computeAngle(SensorModel sm, double positionDistance, double position) { 
		double alfa=0;
		double vertical=0;
		double horizontal=0;
		if((sm.getTrackPosition()>0 && position>0) || (sm.getTrackPosition()<0 && position<0)) { //nasa pozicia aj pozicia kde smerujeme je na rovnakej strane trate
			vertical=Math.abs((trackWidth/2.0)*Math.abs(sm.getTrackPosition())-trackWidth/2.0*Math.abs(position));
			horizontal=positionDistance-sm.getDistanceFromStartLine();
			alfa = Math.atan(vertical/horizontal);
		}
		else {
			vertical=Math.abs((trackWidth/2.0)*Math.abs(sm.getTrackPosition())+trackWidth/2.0*Math.abs(position)); //nasa pozicia a pozicia kde smerujeme su na roznych stranach trate
			horizontal=positionDistance-sm.getDistanceFromStartLine();
			alfa = Math.atan(vertical/horizontal);
		}
		return alfa;
	}
	
	/**
	 * Metoda na upravu smeru jazdy a navigaciu na konkretnu poziciu na trati.
	 * @param action aktualne nastavenie efektorov
	 * @param sm senzory
	 * @param positionDistance pozicia, na ktoru sa chceme dostat - kilometer od zaciatku trate
	 * @param position poloha na trati na danej pozicii
	 * @return upravena action - natocenie kolies
	 * @author Juraj Kosmel
	 */
	private Action navigateToPosition(Action action, SensorModel sm, double positionDistance, double position) {
		double alfa, beta, sum, diff;                                 //uhol natocenia k cielovej pozicii a uhol auta s osou, sucet a rozdiel uhlov
		alfa=computeAngle(sm, positionDistance, position); //vypocet uhla k miestu kde chceme smerovat
		beta = sm.getAngleToTrackAxis();				   //uhol nasho aktualneho smeru
		sum=Math.abs(alfa)+Math.abs(beta);				   //sucet uhlov
		diff=Math.abs(Math.abs(alfa)-Math.abs(beta));				   //rozdiel uhlov
		System.out.println("beta: "+Math.toDegrees(beta)+" alfa: "+Math.toDegrees(alfa));
		
		if(beta>0) { //som otoceny vpravo
			if(sm.getTrackPosition() >= position) { //chcem ist vpravo vzhladom na moju poziciu
				if(Math.abs(beta)<=alfa) {
					action.steering=-(sum/0.36651943);
					System.out.println("natoceny vpravo chcem vpravo beta mensia");
				}
				else {
					action.steering=diff/0.36651943;
					System.out.println("natoceny vpravo chcem vpravo beta vacsia");
				}
			}
			else {								   //chcem ist vlavo vzhladom na moju poziciu
				action.steering=sum/0.36651943;
				System.out.println("natoceny vpravo chcem vlavo");
			}
		}
		else {		 //som otoceny vlavo
			if(sm.getTrackPosition() >= position) { //chcem ist vpravo vzhladom na moju poziciu
				action.steering=-(sum/0.36651943);
				System.out.println("natoceny vlavo chcem vpravo");
			}
			else {								   //chcem ist vlavo vzhladom na moju poziciu
				if(Math.abs(beta)<=alfa) {
					action.steering=sum/0.36651943;
					System.out.println("natoceny vlavo chcem vlavo beta mensia");
				}
				else {
					action.steering=-(diff/0.36651943);
					System.out.println("natoceny vlavo chcem vlavo beta vacsia");
				}
			}
		}
		System.out.println("steering: "+action.steering);
		return action;
	}
	
	
	
	
	
	/**
	 * Metoda urcujuca vztah medzi najdlhsim uhlom a natocenim volantu.
	 * @param longestAngle uhol k najvzdialenejsiemu okraju trate
	 * @return natocenie volantu
	 */
	private double steeringAdjustment(double longestAngle)
	{
		if (Math.abs(longestAngle) > 21)
			if (longestAngle > 0)
				return -1;
			else
				return 1;
		
		return -longestAngle / 21;
	}

	/**
	 * Metoda prebrana zo SimpleDriver sluziaca na navrat na trat.
	 * @param sm senzory
	 * @return akcia smerujuca k nabraniu vhodneho smeru
	 */
	private Action recoverToTrack(SensorModel sm)
	{
		//System.out.println("Prebieha zotavovanie z uviaznutia.");
		Action action = new Action();
		action.accelerate = 1;
					
		// zistime, ci je auto natocene spravnym smerom alebo smeruje von z trate
		if (sm.getAngleToTrackAxis() * sm.getTrackPosition() > 0)
		{
			action.gear = 1;
			action.steering = sm.getAngleToTrackAxis() / (Math.PI / 4);
		}
		else
		{
			action.gear = -1;
			action.steering = -sm.getAngleToTrackAxis() / (Math.PI / 4);
		}
		
		return action;
	}

	/**
	 * Metoda vrati priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 * v rozsahu <-90; 90>. Priemerovana je len jedna suvisla postupnost senzorov.
	 * @param trackEdgeSensors senzory okrajov trate
	 * @return priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 */
	private double computeAvgLongestSensorAngle(double[] trackEdgeSensors)
	{
		int max = 0, numberOfLongestSensors = 1;
		double angle = 0;

		// najde index prveho senzora v smere k najvzdialenejsiemu okraju trate
		for (int i = 1; i < trackEdgeSensors.length; i++)
			if (trackEdgeSensors[max] < trackEdgeSensors[i])
				max = i;
		
		// zisti kolko senzorov ma najvyssiu hodnotu
		while (max + numberOfLongestSensors < trackEdgeSensors.length &&
				trackEdgeSensors[max] == trackEdgeSensors[max + numberOfLongestSensors])
			numberOfLongestSensors++;

		// vypocet priemeru, vysledok je z rozsahu <-90; 90>
		for (int i = 0; i < numberOfLongestSensors; i++)
			angle = max + i;
		angle = (angle / numberOfLongestSensors) * 10 - 90;
		
		return angle;
	}

	/**
	 * Metoda vracajuca vhodnu hodnotu prevodoveho stupna.
	 * 
	 * Implementacia nie je urcena na zaradenie do "produkcneho" kodu.
	 * Jej cielom je uplna oddelenost od ostatnych prvkov ovladaca pre
	 * prehladnost vyvoja.
	 */
	private int setGear(SensorModel sensors)
	{
		int gear = sensors.getGear();

		int[]  gearUp = { 9500, 9400, 9500, 9500, 9500, 0 };
		int[]  gearDown = { 0, 3300, 6200, 7000, 7300, 7700 };
		
		double rpm = sensors.getRPM();
		if (gear < 1)
			gear = 1;

		// check if the RPM value of car is greater than the one suggested
		// to shift up the gear from the current one
		else if (gear < 6 && rpm >= gearUp[gear-1])
			gear++;
		
		// check if the RPM value of car is lower than the one suggested
		// to shift down the gear from the current one
		else if (gear > 1 && rpm <= gearDown[gear-1])
			gear--;
		
		return gear;
	}
	
	public void reset()
	{
		System.out.println("restarting the race");
	}

	public void shutdown()
	{
		System.out.println("shutdown");		
	}
}
