package controller;

import controller.TurnDetector.Direction;

/**
 * Trieda urcena na uchovanie zaznamu o prejdenych usekoch trate. 
 * 
 * @author Ivan Valencik
 */
public class Record
{
	private Direction dir;
	private int duration = 1;
	private double avgAngle, maxAngle, distance, avgSpeed, maxSpeed;
	
	public Record(double steering, double distance, double speed)
	{
		if (steering == 0)
			dir = Direction.STRAIGHT;
		if (steering > 0)
			dir = Direction.LEFT;
		if (steering < 0)
			dir = Direction.RIGHT;
		
		avgAngle = maxAngle = Math.abs(steering);
		avgSpeed = maxSpeed = speed;
		this.distance = distance;
	}

	public void update(double steering, double speed)
	{
		double absSteering = Math.abs(steering);
		
		if (maxAngle < absSteering)
			maxAngle = absSteering;
		avgAngle = rollingAverage(absSteering, avgAngle);
		
		if (maxSpeed < speed)
			maxSpeed = speed;
		avgSpeed = rollingAverage(speed, avgSpeed);

		duration++;
	}
	
	public void merge(Record record)
	{
		avgAngle = (avgAngle * duration + record.getAverageAngle() * record.getDuration()) /
			(duration + record.getDuration());
		if (record.getMaximumAngle() > maxAngle)
			maxAngle = record.getMaximumAngle();
		
		avgSpeed = (avgSpeed * duration + record.getAverageSpeed() * record.getDuration()) /
			(duration + record.getDuration());
		if (record.getMaximumSpeed() > maxSpeed)
			maxSpeed = record.getMaximumSpeed();
		
		duration += record.getDuration();
	}
	
	public Direction getDirection()
	{
		return dir;
	}
	
	public double getAverageAngle()
	{
		return avgAngle;
	}
	
	public double getMaximumAngle()
	{
		return maxAngle;
	}
	
	public int getDuration()
	{
		return duration;
	}
	
	public double getDistance()
	{
		return distance;
	}

	public double getMaximumSpeed()
	{
		return maxSpeed;
	}

	public double getAverageSpeed()
	{
		return avgSpeed;
	}

	private double rollingAverage(double value, double average)
	{
		return (average * duration + value) / (duration + 1);
	}
}
