package controller;

import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;

/**
 * Trieda vytvarajuca zaznamy o senzoroch, podla ktorych sa auto pohybovalo. Bezi pri tom v inom vlakne
 * ako ovladaci modul vozidla.
 * 
 * @author Adam Brcek
 */
public class SensorDetector implements Runnable
{
	private final BlockingQueue<SensorInfo> sensorInfoPipe;
	
	SensorInfo sI = null;
	ArrayList<SensorInfo> sensorList = new ArrayList<SensorInfo>();
	
	public SensorDetector(BlockingQueue<SensorInfo> source)
	{
		sensorInfoPipe = source;
	}
	
	public void run()
	{
		try {
			while (true)
			{
				SensorInfo sInfo = sensorInfoPipe.take();
				update(sInfo);
			}
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri prijmani akcie do modelu trate.");
			e.printStackTrace();
		}
	}
	
	/**
	 * Na zaklade hodnoty senzora urci, ci doplni vytvarany zaznam alebo zacne pracovat s novym.
	 * @param sInfo Zaznam o case, v ktorom sa auto pohzbovalo podla jeneho senzora
	 */
	public void update(SensorInfo sInfo)
	{
		if (sI == null)
			sI = sInfo;
		else
		{
			if (sI.getSensor() != sInfo.getSensor())
			{
				sensorList.add(sI);
				System.out.println("Zaznam cislo "+(sensorList.size()+1));
				System.out.println("Senzor "+sI.getSensor());
				System.out.println("Trvanie "+sI.getDuration()+" tickov");
				System.out.println("Zaciatok useku "+sI.getStart()+" metrov");
				System.out.println();
				sI = sInfo;
			}
			else
			{
				sI.setDuration(sI.getDuration()+1);
			}
		}
	}
}
